#!/usr/bin/env python

"""
A Web interface to a calendar event.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import date, timedelta
from imiptools.data import get_uri, uri_dict, uri_values
from imiptools.dates import format_datetime, get_datetime_item, \
                            to_date, to_timezone
from imiptools.mail import Messenger
from imiptools.period import have_conflict
from imipweb.data import EventPeriod, \
                         event_period_from_period, form_period_from_period, \
                         FormDate, FormPeriod, PeriodError
from imipweb.client import ManagerClient
from imipweb.resource import Resource
import pytz

class EventPage(Resource):

    "A request handler for the event page."

    def __init__(self, resource=None, messenger=None):
        Resource.__init__(self, resource)
        self.messenger = messenger or Messenger()

    # Various property values and labels.

    property_items = [
        ("SUMMARY", "Summary"),
        ("DTSTART", "Start"),
        ("DTEND", "End"),
        ("ORGANIZER", "Organiser"),
        ("ATTENDEE", "Attendee"),
        ]

    partstat_items = [
        ("NEEDS-ACTION", "Not confirmed"),
        ("ACCEPTED", "Attending"),
        ("TENTATIVE", "Tentatively attending"),
        ("DECLINED", "Not attending"),
        ("DELEGATED", "Delegated"),
        (None, "Not indicated"),
        ]

    # Access to stored object information.

    def is_organiser(self, obj):
        return get_uri(obj.get_value("ORGANIZER")) == self.user

    def get_stored_attendees(self, obj):
        return uri_values(obj.get_values("ATTENDEE") or [])

    def get_stored_main_period(self, obj):

        """
        Return the main event period for the given 'obj'.
        """

        dtstart, dtstart_attr = obj.get_datetime_item("DTSTART")

        if obj.has_key("DTEND"):
            dtend, dtend_attr = obj.get_datetime_item("DTEND")
        elif obj.has_key("DURATION"):
            duration = obj.get_duration("DURATION")
            dtend = dtstart + duration
            dtend_attr = dtstart_attr
        else:
            dtend, dtend_attr = dtstart, dtstart_attr

        return EventPeriod(dtstart, dtend, self.get_tzid(), None, dtstart_attr, dtend_attr)

    def get_stored_recurrences(self, obj):

        "Return recurrences computed using the given 'obj'."

        recurrences = []
        for period in self.get_periods(obj):
            if period.origin != "DTSTART":
                recurrences.append(period)
        return recurrences

    # Request logic methods.

    def is_initial_load(self):

        "Return whether the event is being loaded and shown for the first time."

        return not self.env.get_args().has_key("editing")

    def handle_request(self, obj):

        """
        Handle actions involving the given 'obj' as an object's representation,
        returning an error if one occurred, or None if the request was
        successfully handled.
        """

        # Handle a submitted form.

        args = self.env.get_args()
        uid = obj.get_uid()
        recurrenceid = obj.get_recurrenceid()

        # Get the possible actions.

        reply = args.has_key("reply")
        discard = args.has_key("discard")
        create = args.has_key("create")
        cancel = args.has_key("cancel")
        ignore = args.has_key("ignore")
        save = args.has_key("save")

        have_action = reply or discard or create or cancel or ignore or save

        if not have_action:
            return ["action"]

        # If ignoring the object, return to the calendar.

        if ignore:
            self.redirect(self.env.get_path())
            return None

        # Update the object.

        single_user = False

        if reply or create or cancel or save:

            # Update principal event details if organiser.

            if self.is_organiser(obj):

                # Update time periods (main and recurring).

                try:
                    period = self.handle_main_period()
                except PeriodError, exc:
                    return exc.args

                try:
                    periods = self.handle_recurrence_periods()
                except PeriodError, exc:
                    return exc.args

                # Set the periods in the object, first obtaining removed and
                # modified period information.

                to_unschedule = self.get_removed_periods()

                obj.set_period(period)
                obj.set_periods(periods)

                # Update summary.

                if args.has_key("summary"):
                    obj["SUMMARY"] = [(args["summary"][0], {})]

                # Obtain any participants and those to be removed.

                attendees = self.get_attendees_from_page()
                removed = [attendees[int(i)] for i in args.get("remove", [])]
                to_cancel = self.update_attendees(obj, attendees, removed)
                single_user = not attendees or attendees == [self.user]

            # Update attendee participation for the current user.

            if args.has_key("partstat"):
                self.update_participation(obj, args["partstat"][0])

        # Process any action.

        invite = not save and create and not single_user
        save = save or create and single_user

        handled = True

        if reply or invite or cancel:

            client = ManagerClient(obj, self.user, self.messenger)

            # Process the object and remove it from the list of requests.

            if reply and client.process_received_request():
                self.remove_request(uid, recurrenceid)

            elif self.is_organiser(obj) and (invite or cancel):

                # Invitation, uninvitation and unscheduling...

                if client.process_created_request(
                    invite and "REQUEST" or "CANCEL", to_cancel, to_unschedule):

                    self.remove_request(uid, recurrenceid)

        # Save single user events.

        elif save:
            self.store.set_event(self.user, uid, recurrenceid, node=obj.to_node())
            self.update_freebusy(uid, recurrenceid, obj)
            self.remove_request(uid, recurrenceid)

        # Remove the request and the object.

        elif discard:
            self.remove_from_freebusy(uid, recurrenceid)
            self.remove_event(uid, recurrenceid)
            self.remove_request(uid, recurrenceid)

        else:
            handled = False

        # Upon handling an action, redirect to the main page.

        if handled:
            self.redirect(self.env.get_path())

        return None

    def handle_main_period(self):

        "Return period details for the main start/end period in an event."

        return self.get_main_period().as_event_period()

    def handle_recurrence_periods(self):

        "Return period details for the recurrences specified for an event."

        return [p.as_event_period(i) for i, p in enumerate(self.get_recurrences())]

    def get_date_control_values(self, name, multiple=False, tzid_name=None):

        """
        Return a dictionary containing date, time and tzid entries for fields
        starting with 'name'. If 'multiple' is set to a true value, many
        dictionaries will be returned corresponding to a collection of
        datetimes. If 'tzid_name' is specified, the time zone information will
        be acquired from a field starting with 'tzid_name' instead of 'name'.
        """

        args = self.env.get_args()

        dates = args.get("%s-date" % name, [])
        hours = args.get("%s-hour" % name, [])
        minutes = args.get("%s-minute" % name, [])
        seconds = args.get("%s-second" % name, [])
        tzids = args.get("%s-tzid" % (tzid_name or name), [])

        # Handle absent values by employing None values.

        field_values = map(None, dates, hours, minutes, seconds, tzids)

        if not field_values and not multiple:
            all_values = FormDate()
        else:
            all_values = []
            for date, hour, minute, second, tzid in field_values:
                value = FormDate(date, hour, minute, second, tzid or self.get_tzid())

                # Return a single value or append to a collection of all values.

                if not multiple:
                    return value
                else:
                    all_values.append(value)

        return all_values

    def get_current_main_period(self, obj):

        """
        Return the currently active main period for 'obj' depending on whether
        editing has begun or whether the object has just been loaded.
        """

        if self.is_initial_load() or not self.is_organiser(obj):
            return self.get_stored_main_period(obj)
        else:
            return self.get_main_period()

    def get_main_period(self):

        "Return the main period defined in the event form."

        args = self.env.get_args()

        dtend_enabled = args.get("dtend-control", [None])[0]
        dttimes_enabled = args.get("dttimes-control", [None])[0]
        start = self.get_date_control_values("dtstart")
        end = self.get_date_control_values("dtend")

        return FormPeriod(start, end, dtend_enabled, dttimes_enabled, self.get_tzid())

    def get_current_recurrences(self, obj):

        """
        Return recurrences for 'obj' using the original object where no editing
        is in progress, using form data otherwise.
        """

        if self.is_initial_load() or not self.is_organiser(obj):
            return self.get_stored_recurrences(obj)
        else:
            return self.get_recurrences()

    def get_recurrences(self):

        "Return the recurrences defined in the event form."

        args = self.env.get_args()

        all_dtend_enabled = args.get("dtend-control-recur", [])
        all_dttimes_enabled = args.get("dttimes-control-recur", [])
        all_starts = self.get_date_control_values("dtstart-recur", multiple=True)
        all_ends = self.get_date_control_values("dtend-recur", multiple=True, tzid_name="dtstart-recur")
        all_origins = args.get("recur-origin", [])

        periods = []

        for index, (start, end, dtend_enabled, dttimes_enabled, origin) in \
            enumerate(map(None, all_starts, all_ends, all_dtend_enabled, all_dttimes_enabled, all_origins)):

            dtend_enabled = str(index) in all_dtend_enabled
            dttimes_enabled = str(index) in all_dttimes_enabled
            period = FormPeriod(start, end, dtend_enabled, dttimes_enabled, self.get_tzid(), origin)
            periods.append(period)

        return periods

    def get_removed_periods(self):

        "Return a list of recurrence periods to remove upon updating an event."

        to_unschedule = []
        args = self.env.get_args()
        for i in args.get("recur-remove", []):
            to_unschedule.append(periods[int(i)])
        return to_unschedule

    def get_current_attendees(self, obj):

        """
        Return attendees for 'obj' depending on whether the object is being
        edited.
        """

        if self.is_initial_load() or not self.is_organiser(obj):
            return self.get_stored_attendees(obj)
        else:
            return self.get_attendees_from_page()

    def get_attendees_from_page(self):

        """
        Return attendees from the request, normalised for iCalendar purposes,
        and without duplicates.
        """

        args = self.env.get_args()

        attendees = args.get("attendee", [])
        unique_attendees = set()
        ordered_attendees = []

        for attendee in attendees:
            if not attendee.strip():
                continue
            attendee = get_uri(attendee)
            if attendee not in unique_attendees:
                unique_attendees.add(attendee)
                ordered_attendees.append(attendee)

        return ordered_attendees

    def update_attendees_from_page(self, obj):

        "Add or remove attendees. This does not affect the stored object."

        args = self.env.get_args()

        attendees = self.get_attendees_from_page()
        existing_attendees = self.get_stored_attendees(obj)

        if args.has_key("add"):
            attendees.append("")

        # Only actually remove attendees if the event is unsent, if the attendee
        # is new, or if it is the current user being removed.

        if args.has_key("remove"):
            for i in args["remove"]:
                try:
                    attendee = attendees[int(i)]
                except IndexError:
                    continue

                existing = attendee in existing_attendees

                if not existing or not obj.is_shared() or attendee == self.user:
                    attendees.remove(attendee)

        return attendees

    # Page fragment methods.

    def show_request_controls(self, obj):

        "Show form controls for a request concerning 'obj'."

        page = self.page
        args = self.env.get_args()

        attendees = self.get_current_attendees(obj)
        is_attendee = self.user in attendees
        is_request = (obj.get_uid(), obj.get_recurrenceid()) in self._get_requests()

        # Show appropriate options depending on the role of the user.

        if is_attendee and not self.is_organiser(obj):
            page.p("An action is required for this request:")

            page.p()
            self._control("reply", "submit", "Send reply")
            page.add(" ")
            self._control("discard", "submit", "Discard event")
            page.add(" ")
            self._control("ignore", "submit", "Do nothing for now")
            page.p.close()

        if self.is_organiser(obj):
            page.p("As organiser, you can perform the following:")

            page.p()
            self._control("create", "submit", not obj.is_shared() and "Create event" or "Update event")
            page.add(" ")

            if obj.is_shared() and not is_request:
                self._control("cancel", "submit", "Cancel event")
            else:
                self._control("discard", "submit", "Discard event")

            page.add(" ")
            self._control("save", "submit", "Save without sending")
            page.p.close()

    def show_object_on_page(self, obj, errors=None):

        """
        Show the calendar object with the representation 'obj' on the current
        page. If 'errors' is given, show a suitable message for the different
        errors provided.
        """

        page = self.page
        page.form(method="POST")

        # Add a hidden control to help determine whether editing has already begun.

        self._control("editing", "hidden", "true")

        uid = obj.get_uid()
        args = self.env.get_args()

        # Obtain basic event information, generating any necessary editing controls.

        if self.is_initial_load() or not self.is_organiser(obj):
            attendees = self.get_stored_attendees(obj)
        else:
            attendees = self.update_attendees_from_page(obj)

        p = self.get_current_main_period(obj)
        self.show_object_datetime_controls(p)

        # Obtain any separate recurrences for this event.

        recurrenceid = obj.get_recurrenceid()
        recurrenceids = self._get_active_recurrences(uid)
        replaced = not recurrenceid and p.is_replaced(recurrenceids)

        # Provide a summary of the object.

        page.table(class_="object", cellspacing=5, cellpadding=5)
        page.thead()
        page.tr()
        page.th("Event", class_="mainheading", colspan=2)
        page.tr.close()
        page.thead.close()
        page.tbody()

        for name, label in self.property_items:
            field = name.lower()

            items = obj.get_items(name) or []
            rowspan = len(items)

            if name == "ATTENDEE":
                rowspan = len(attendees) + 1 # for the add button
            elif not items:
                continue

            page.tr()
            page.th(label, class_="objectheading %s%s" % (field, errors and field in errors and " error" or ""), rowspan=rowspan)

            # Handle datetimes specially.

            if name in ["DTSTART", "DTEND"]:
                if not replaced:

                    # Obtain the datetime.

                    is_start = name == "DTSTART"

                    # Where no end datetime exists, use the start datetime as the
                    # basis of any potential datetime specified if dt-control is
                    # set.

                    self.show_datetime_controls(obj, is_start and p.get_form_start() or p.get_form_end(), is_start)

                elif name == "DTSTART":
                    page.td(class_="objectvalue %s replaced" % field, rowspan=2)
                    page.a("First occurrence replaced by a separate event", href=self.link_to(uid, replaced))
                    page.td.close()

                page.tr.close()

            # Handle the summary specially.

            elif name == "SUMMARY":
                value = args.get("summary", [obj.get_value(name)])[0]

                page.td(class_="objectvalue summary")
                if self.is_organiser(obj):
                    self._control("summary", "text", value, size=80)
                else:
                    page.add(value)
                page.td.close()
                page.tr.close()

            # Handle attendees specially.

            elif name == "ATTENDEE":
                attendee_map = dict(items)
                first = True

                for i, value in enumerate(attendees):
                    if not first:
                        page.tr()
                    else:
                        first = False

                    # Obtain details of attendees to supply attributes.

                    self.show_attendee(obj, i, value, attendee_map.get(value))
                    page.tr.close()

                # Allow more attendees to be specified.

                if self.is_organiser(obj):
                    if not first:
                        page.tr()

                    page.td()
                    self._control("add", "submit", "add", id="add", class_="add")
                    page.label("Add attendee", for_="add", class_="add")
                    page.td.close()
                    page.tr.close()

            # Handle potentially many values of other kinds.

            else:
                first = True

                for i, (value, attr) in enumerate(items):
                    if not first:
                        page.tr()
                    else:
                        first = False

                    page.td(class_="objectvalue %s" % field)
                    page.add(value)
                    page.td.close()
                    page.tr.close()

        page.tbody.close()
        page.table.close()

        self.show_recurrences(obj, errors)
        self.show_conflicting_events(obj)
        self.show_request_controls(obj)

        page.form.close()

    def show_attendee(self, obj, i, attendee, attendee_attr):

        """
        For the given object 'obj', show the attendee in position 'i' with the
        given 'attendee' value, having 'attendee_attr' as any stored attributes.
        """

        page = self.page
        args = self.env.get_args()

        existing = attendee_attr is not None
        partstat = attendee_attr and attendee_attr.get("PARTSTAT")

        page.td(class_="objectvalue")

        # Show a form control as organiser for new attendees.

        if self.is_organiser(obj) and (not existing or not obj.is_shared()):
            self._control("attendee", "value", attendee, size="40")
        else:
            self._control("attendee", "hidden", attendee)
            page.add(attendee)
        page.add(" ")

        # Show participation status, editable for the current user.

        if attendee == self.user:
            self._show_menu("partstat", partstat, self.partstat_items, "partstat")

        # Allow the participation indicator to act as a submit
        # button in order to refresh the page and show a control for
        # the current user, if indicated.

        elif self.is_organiser(obj) and not existing:
            self._control("partstat-refresh", "submit", "refresh", id="partstat-%d" % i, class_="refresh")
            page.label(dict(self.partstat_items).get(partstat, ""), for_="partstat-%s" % i, class_="partstat")
        else:
            page.span(dict(self.partstat_items).get(partstat, ""), class_="partstat")

        # Permit organisers to remove attendees.

        if self.is_organiser(obj):

            # Permit the removal of newly-added attendees.

            remove_type = (not existing or not obj.is_shared() or attendee == self.user) and "submit" or "checkbox"
            self._control("remove", remove_type, str(i), str(i) in args.get("remove", []), id="remove-%d" % i, class_="remove")

            page.label("Remove", for_="remove-%d" % i, class_="remove")
            page.label("Uninvited", for_="remove-%d" % i, class_="removed")

        page.td.close()

    def show_recurrences(self, obj, errors=None):

        """
        Show recurrences for the object having the given representation 'obj'.
        If 'errors' is given, show a suitable message for the different errors
        provided.
        """

        page = self.page

        # Obtain any parent object if this object is a specific recurrence.

        uid = obj.get_uid()
        recurrenceid = obj.get_recurrenceid()

        if recurrenceid:
            parent = self.get_stored_object(uid, None)
            if not parent:
                return

            page.p()
            page.a("This event modifies a recurring event.", href=self.link_to(uid))
            page.p.close()

        # Obtain the periods associated with the event.
        # NOTE: Add a control to add recurrences here.

        recurrences = self.get_current_recurrences(obj)

        if len(recurrences) < 1:
            return

        recurrenceids = self._get_recurrences(uid)

        page.p("This event occurs on the following occasions within the next %d days:" % self.get_window_size())

        # Show each recurrence in a separate table if editable.

        if self.is_organiser(obj) and recurrences:

            for index, period in enumerate(recurrences):
                self.show_recurrence(obj, index, period, recurrenceid, recurrenceids, errors)

        # Otherwise, use a compact single table.

        else:
            page.table(cellspacing=5, cellpadding=5, class_="recurrence")
            page.caption("Occurrences")
            page.thead()
            page.tr()
            page.th("Start", class_="objectheading start")
            page.th("End", class_="objectheading end")
            page.tr.close()
            page.thead.close()
            page.tbody()

            for index, period in enumerate(recurrences):
                page.tr()
                self.show_recurrence_label(period, recurrenceid, recurrenceids, True)
                self.show_recurrence_label(period, recurrenceid, recurrenceids, False)
                page.tr.close()

            page.tbody.close()
            page.table.close()

    def show_recurrence(self, obj, index, period, recurrenceid, recurrenceids, errors=None):

        """
        Show recurrence controls for a recurrence provided by 'obj' with the
        given 'index' position in the list of periods, the given 'period'
        details, where a 'recurrenceid' indicates any specific recurrence, and
        where 'recurrenceids' indicates all known additional recurrences for the
        object.

        If 'errors' is given, show a suitable message for the different errors
        provided.
        """

        page = self.page
        args = self.env.get_args()

        p = event_period_from_period(period)
        replaced = not recurrenceid and p.is_replaced(recurrenceids)

        # Isolate the controls from neighbouring tables.

        page.div()

        self.show_object_datetime_controls(period, index)

        page.table(cellspacing=5, cellpadding=5, class_="recurrence")
        page.caption(period.origin == "RRULE" and "Occurrence from rule" or "Occurrence")
        page.tbody()

        page.tr()
        error = errors and ("dtstart", index) in errors and " error" or ""
        page.th("Start", class_="objectheading start%s" % error)
        self.show_recurrence_controls(obj, index, period, recurrenceid, recurrenceids, True)
        page.tr.close()
        page.tr()
        error = errors and ("dtend", index) in errors and " error" or ""
        page.th("End", class_="objectheading end%s" % error)
        self.show_recurrence_controls(obj, index, period, recurrenceid, recurrenceids, False)
        page.tr.close()

        # Permit the removal of recurrences.

        if not replaced:
            page.tr()
            page.th("")
            page.td()

            remove_type = not obj.is_shared() or not period.origin and "submit" or "checkbox"
            self._control("recur-remove", remove_type, str(index),
                str(index) in args.get("recur-remove", []),
                id="recur-remove-%d" % index, class_="remove")

            page.label("Remove", for_="recur-remove-%d" % index, class_="remove")
            page.label("Removed", for_="recur-remove-%d" % index, class_="removed")

            page.td.close()
            page.tr.close()

        page.tbody.close()
        page.table.close()

        page.div.close()

    def show_conflicting_events(self, obj):

        """
        Show conflicting events for the object having the representation 'obj'.
        """

        page = self.page
        uid = obj.get_uid()
        recurrenceid = obj.get_recurrenceid()
        recurrenceids = self._get_active_recurrences(uid)

        # Obtain the user's timezone.

        tzid = self.get_tzid()
        periods = self.get_periods(obj)

        # Indicate whether there are conflicting events.

        conflicts = []
        attendee_map = uri_dict(obj.get_value_map("ATTENDEE"))

        for participant in self.get_current_attendees(obj):
            if participant == self.user:
                freebusy = self.store.get_freebusy(participant)
            else:
                freebusy = self.store.get_freebusy_for_other(self.user, participant)

            if not freebusy:
                continue

            # Obtain any time zone details from the suggested event.

            _dtstart, attr = obj.get_item("DTSTART")
            tzid = attr.get("TZID", tzid)

            # Show any conflicts with periods of actual attendance.

            participant_attr = attendee_map.get(participant)
            partstat = participant_attr and participant_attr.get("PARTSTAT")
            recurrences = obj.get_recurrence_start_points(recurrenceids, tzid)

            for p in have_conflict(freebusy, periods, True):
                if not recurrenceid and p.is_replaced(recurrences):
                    continue

                if ( # Unidentified or different event
                     (p.uid != uid or recurrenceid and p.recurrenceid and p.recurrenceid != recurrenceid) and
                     # Different period or unclear participation with the same period
                     (p not in periods or not partstat in ("ACCEPTED", "TENTATIVE")) and
                     # Participant not limited to organising
                     p.transp != "ORG"
                   ):

                    conflicts.append(p)

        conflicts.sort()

        # Show any conflicts with periods of actual attendance.

        if conflicts:
            page.p("This event conflicts with others:")

            page.table(cellspacing=5, cellpadding=5, class_="conflicts")
            page.thead()
            page.tr()
            page.th("Event")
            page.th("Start")
            page.th("End")
            page.tr.close()
            page.thead.close()
            page.tbody()

            for p in conflicts:

                # Provide details of any conflicting event.

                start = self.format_datetime(to_timezone(p.get_start(), tzid), "long")
                end = self.format_datetime(to_timezone(p.get_end(), tzid), "long")

                page.tr()

                # Show the event summary for the conflicting event.

                page.td()
                if p.summary:
                    page.a(p.summary, href=self.link_to(p.uid, p.recurrenceid))
                else:
                    page.add("(Unspecified event)")
                page.td.close()

                page.td(start)
                page.td(end)

                page.tr.close()

            page.tbody.close()
            page.table.close()

    # Generation of controls within page fragments.

    def show_object_datetime_controls(self, period, index=None):

        """
        Show datetime-related controls if already active or if an object needs
        them for the given 'period'. The given 'index' is used to parameterise
        individual controls for dynamic manipulation.
        """

        p = form_period_from_period(period)

        page = self.page
        args = self.env.get_args()
        _id = self.element_identifier
        _name = self.element_name
        _enable = self.element_enable

        # Add a dynamic stylesheet to permit the controls to modify the display.
        # NOTE: The style details need to be coordinated with the static
        # NOTE: stylesheet.

        if index is not None:
            page.style(type="text/css")

            # Unlike the rules for object properties, these affect recurrence
            # properties.

            page.add("""\
input#dttimes-enable-%(index)d,
input#dtend-enable-%(index)d,
input#dttimes-enable-%(index)d:not(:checked) ~ .recurrence td.objectvalue .time.enabled,
input#dttimes-enable-%(index)d:checked ~ .recurrence td.objectvalue .time.disabled,
input#dtend-enable-%(index)d:not(:checked) ~ .recurrence td.objectvalue.dtend .dt.enabled,
input#dtend-enable-%(index)d:checked ~ .recurrence td.objectvalue.dtend .dt.disabled {
    display: none;
}""" % {"index" : index})

            page.style.close()

        self._control(
            _name("dtend-control", "recur", index), "checkbox",
            _enable(index), p.end_enabled,
            id=_id("dtend-enable", index)
            )

        self._control(
            _name("dttimes-control", "recur", index), "checkbox",
            _enable(index), p.times_enabled,
            id=_id("dttimes-enable", index)
            )

    def show_datetime_controls(self, obj, formdate, show_start):

        """
        Show datetime details from the given 'obj' for the 'formdate', showing
        start details if 'show_start' is set to a true value. Details will
        appear as controls for organisers and labels for attendees.
        """

        page = self.page

        # Show controls for editing as organiser.

        if self.is_organiser(obj):
            page.td(class_="objectvalue dt%s" % (show_start and "start" or "end"))

            if show_start:
                page.div(class_="dt enabled")
                self._show_date_controls("dtstart", formdate)
                page.br()
                page.label("Specify times", for_="dttimes-enable", class_="time disabled enable")
                page.label("Specify dates only", for_="dttimes-enable", class_="time enabled disable")
                page.div.close()

            else:
                page.div(class_="dt disabled")
                page.label("Specify end date", for_="dtend-enable", class_="enable")
                page.div.close()
                page.div(class_="dt enabled")
                self._show_date_controls("dtend", formdate)
                page.br()
                page.label("End on same day", for_="dtend-enable", class_="disable")
                page.div.close()

            page.td.close()

        # Show a label as attendee.

        else:
            dt = formdate.as_datetime()
            if dt:
                page.td(self.format_datetime(dt, "full"))
            else:
                page.td("(Unrecognised date)")

    def show_recurrence_controls(self, obj, index, period, recurrenceid, recurrenceids, show_start):

        """
        Show datetime details from the given 'obj' for the recurrence having the
        given 'index', with the recurrence period described by 'period',
        indicating a start, end and origin of the period from the event details,
        employing any 'recurrenceid' and 'recurrenceids' for the object to
        configure the displayed information.

        If 'show_start' is set to a true value, the start details will be shown;
        otherwise, the end details will be shown.
        """

        page = self.page
        _id = self.element_identifier
        _name = self.element_name

        p = event_period_from_period(period)
        replaced = not recurrenceid and p.is_replaced(recurrenceids)

        # Show controls for editing as organiser.

        if self.is_organiser(obj) and not replaced:
            page.td(class_="objectvalue dt%s" % (show_start and "start" or "end"))

            read_only = period.origin == "RRULE"

            if show_start:
                page.div(class_="dt enabled")
                self._show_date_controls(_name("dtstart", "recur", index), p.get_form_start(), index=index, read_only=read_only)
                if not read_only:
                    page.br()
                    page.label("Specify times", for_=_id("dttimes-enable", index), class_="time disabled enable")
                    page.label("Specify dates only", for_=_id("dttimes-enable", index), class_="time enabled disable")
                page.div.close()

                # Put the origin somewhere.

                self._control("recur-origin", "hidden", p.origin or "")

            else:
                page.div(class_="dt disabled")
                if not read_only:
                    page.label("Specify end date", for_=_id("dtend-enable", index), class_="enable")
                page.div.close()
                page.div(class_="dt enabled")
                self._show_date_controls(_name("dtend", "recur", index), p.get_form_end(), index=index, show_tzid=False, read_only=read_only)
                if not read_only:
                    page.br()
                    page.label("End on same day", for_=_id("dtend-enable", index), class_="disable")
                page.div.close()

            page.td.close()

        # Show label as attendee.

        else:
            self.show_recurrence_label(p, recurrenceid, recurrenceids, show_start)

    def show_recurrence_label(self, period, recurrenceid, recurrenceids, show_start):

        """
        Show datetime details for the given 'period', employing any
        'recurrenceid' and 'recurrenceids' for the object to configure the
        displayed information.

        If 'show_start' is set to a true value, the start details will be shown;
        otherwise, the end details will be shown.
        """

        page = self.page

        p = event_period_from_period(period)
        replaced = not recurrenceid and p.is_replaced(recurrenceids)

        css = " ".join([
            replaced and "replaced" or "",
            p.is_affected(recurrenceid) and "affected" or ""
            ])

        formdate = show_start and p.get_form_start() or p.get_form_end()
        dt = formdate.as_datetime()
        if dt:
            page.td(self.format_datetime(dt, "long"), class_=css)
        else:
            page.td("(Unrecognised date)")

    # Full page output methods.

    def show(self, path_info):

        "Show an object request using the given 'path_info' for the current user."

        uid, recurrenceid = self._get_identifiers(path_info)
        obj = self.get_stored_object(uid, recurrenceid)

        if not obj:
            return False

        errors = self.handle_request(obj)

        if not errors:
            return True

        self.new_page(title="Event")
        self.show_object_on_page(obj, errors)

        return True

    # Utility methods.

    def _control(self, name, type, value, selected=False, **kw):

        """
        Show a control with the given 'name', 'type' and 'value', with
        'selected' indicating whether it should be selected (checked or
        equivalent), and with keyword arguments setting other properties.
        """

        page = self.page
        if selected:
            page.input(name=name, type=type, value=value, checked=selected, **kw)
        else:
            page.input(name=name, type=type, value=value, **kw)

    def _show_menu(self, name, default, items, class_="", index=None):

        """
        Show a select menu having the given 'name', set to the given 'default',
        providing the given (value, label) 'items', and employing the given CSS
        'class_' if specified.
        """

        page = self.page
        values = self.env.get_args().get(name, [default])
        if index is not None:
            values = values[index:]
            values = values and values[0:1] or [default]

        page.select(name=name, class_=class_)
        for v, label in items:
            if v is None:
                continue
            if v in values:
                page.option(label, value=v, selected="selected")
            else:
                page.option(label, value=v)
        page.select.close()

    def _show_date_controls(self, name, default, index=None, show_tzid=True, read_only=False):

        """
        Show date controls for a field with the given 'name' and 'default' form
        date value.

        If 'index' is specified, default field values will be overridden by the
        element from a collection of existing form values with the specified
        index; otherwise, field values will be overridden by a single form
        value.

        If 'show_tzid' is set to a false value, the time zone menu will not be
        provided.

        If 'read_only' is set to a true value, the controls will be hidden and
        labels will be employed instead.
        """

        page = self.page

        # Show dates for up to one week around the current date.

        dt = default.as_datetime()
        if not dt:
            dt = date.today()

        base = to_date(dt)

        # Show a date label with a hidden field if read-only.

        if read_only:
            self._control("%s-date" % name, "hidden", format_datetime(base))
            page.span(self.format_date(base, "long"))

        # Show dates for up to one week around the current date.
        # NOTE: Support paging to other dates.

        else:
            items = []
            for i in range(-7, 8):
                d = base + timedelta(i)
                items.append((format_datetime(d), self.format_date(d, "full")))
            self._show_menu("%s-date" % name, format_datetime(base), items, index=index)

        # Show time details.

        page.span(class_="time enabled")

        if read_only:
            page.span("%s:%s:%s" % (default.get_hour(), default.get_minute(), default.get_second()))
            self._control("%s-hour" % name, "hidden", default.get_hour())
            self._control("%s-minute" % name, "hidden", default.get_minute())
            self._control("%s-second" % name, "hidden", default.get_second())
        else:
            self._control("%s-hour" % name, "text", default.get_hour(), maxlength=2, size=2)
            page.add(":")
            self._control("%s-minute" % name, "text", default.get_minute(), maxlength=2, size=2)
            page.add(":")
            self._control("%s-second" % name, "text", default.get_second(), maxlength=2, size=2)

        # Show time zone details.

        if show_tzid:
            page.add(" ")
            tzid = default.get_tzid() or self.get_tzid()

            # Show a label if read-only or a menu otherwise.

            if read_only:
                self._control("%s-tzid" % name, "hidden", tzid)
                page.span(tzid)
            else:
                self._show_timezone_menu("%s-tzid" % name, tzid, index)

        page.span.close()

    def _show_timezone_menu(self, name, default, index=None):

        """
        Show timezone controls using a menu with the given 'name', set to the
        given 'default' unless a field of the given 'name' provides a value.
        """

        entries = [(tzid, tzid) for tzid in pytz.all_timezones]
        self._show_menu(name, default, entries, index=index)

# vim: tabstop=4 expandtab shiftwidth=4
